﻿//-----------------------------------------------------------------------------
// Game1.cs
//
// Kyoung Shin Park @ Dankook University (4/4/2011)
//-----------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;


namespace Primitives3D
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;

        // FPS
        static FpsCounter fpsCounter;

        // input
        GamePadState currentGamePadState;
        GamePadState prevGamePadState;
        KeyboardState currentKeyboardState;
        KeyboardState prevKeyboardState;

        // The size of the world. The world is a bounding box ranging from -worldSize to worldSize on
        // the X and Z axis, and from 0 to worldSize on the Y axis.
        const float worldSize = 20f;

        // The maximum number of spheres in our world
        const int maximumNumberOfSpheres = 200;

        // An array of spheres and the number of currently active spheres
        Sphere[] spheres = new Sphere[maximumNumberOfSpheres];
        int activeSphereCount = 50;

        // Are we colliding the spheres against each other?
        bool collideSpheres = true;
 
        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            base.Initialize();

            // fpsCounter game component
            fpsCounter = new FpsCounter(this);
            Components.Add(fpsCounter);
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>("hudfont");

            // TODO: use this.Content to load your game content here

            // Create our spheres
            CreateSpheres();
        }

        /// <summary>
        /// Helper method that creates all of our spheres.
        /// </summary>
        private void CreateSpheres()
        {
            // Create a random number generator
            Random random = new Random();

            // These are the various colors we use when creating the spheres
            Color[] sphereColors = new[]
            {
                Color.Red,
                Color.Blue,
                Color.Green,
                Color.Orange,
                Color.Pink,
                Color.Purple,
                Color.Yellow
            };

            // The radius of a sphere
            const float radius = 1f;

            for (int i = 0; i < maximumNumberOfSpheres; i++)
            {
                // Create the sphere
                Sphere sphere = new Sphere(GraphicsDevice, radius);

                // Position the sphere in our world
                sphere.Position = new Vector3(
                    RandomFloat(random, -worldSize + radius, worldSize - radius),
                    RandomFloat(random, radius, worldSize - radius),
                    RandomFloat(random, -worldSize + radius, worldSize - radius));

                // Pick a random color for the sphere
                sphere.Color = sphereColors[random.Next(sphereColors.Length)];

                // Create a random velocity vector
                sphere.Velocity = new Vector3(
                    RandomFloat(random, -10f, 10f),
                    RandomFloat(random, -10f, 10f),
                    RandomFloat(random, -10f, 10f));

                // Add the sphere to our array
                spheres[i] = sphere;
            }
        }

        /// <summary>
        /// A helper method that generates a random float in a given range.
        /// </summary>
        private static float RandomFloat(Random random, float min, float max)
        {
            return (float)random.NextDouble() * (max - min) + min;
        }


        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // TODO: Add your update logic here
            HandleInput();

            // Update all of the spheres in the game, handling collision if desired
            UpdateSpheres(gameTime);

            base.Update(gameTime);
        }
        /// <summary>
        /// Helper method that updates our sphere simulation.
        /// </summary>
        private void UpdateSpheres(GameTime gameTime)
        {
            // Update all spheres and perform collision against the world
            for (int i = 0; i < activeSphereCount; i++)
            {
                Sphere s = spheres[i];
                s.Update(gameTime);
                BounceSphereInWorld(s);
            }

            // If we are colliding spheres against each other
            if (collideSpheres)
            {
                // Iterate over the list twice to compare the spheres against each other
                for (int i = 0; i < activeSphereCount; i++)
                {
                    for (int j = 0; j < activeSphereCount; j++)
                    {
                        // Make sure we don't collid a sphere with itself
                        if (i == j)
                            continue;

                        // Get the spheres
                        Sphere a = spheres[i];
                        Sphere b = spheres[j];

                        // If the spheres are intersecting
                        if (a.Bounds.Intersects(b.Bounds))
                        {
                            // Get the vector between their centers
                            Vector3 delta = b.Position - a.Position;

                            // Calculate the point halfway between the spheres
                            Vector3 center = a.Position + delta / 2f;

                            // Normalize the delta vector
                            delta.Normalize();

                            // Move the spheres to resolve the collision
                            a.Position = center - delta * a.Radius;
                            b.Position = center + delta * b.Radius;

                            // Reflect the velocities to bounce the spheres
                            a.Velocity = Vector3.Normalize(Vector3.Reflect(a.Velocity, delta)) * b.Velocity.Length();
                            b.Velocity = Vector3.Normalize(Vector3.Reflect(b.Velocity, delta)) * a.Velocity.Length();
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Helper method that keeps a sphere in the world by bouncing it off the walls.
        /// </summary>
        private static void BounceSphereInWorld(Sphere s)
        {
            // First test along the X axis, flipping the velocity if a collision occurs.
            if (s.Position.X < -worldSize + s.Radius)
            {
                s.Position.X = -worldSize + s.Radius;
                if (s.Velocity.X < 0f)
                    s.Velocity.X *= -1f;
            }
            else if (s.Position.X > worldSize - s.Radius)
            {
                s.Position.X = worldSize - s.Radius;
                if (s.Velocity.X > 0f)
                    s.Velocity.X *= -1f;
            }

            // Then we test the Y axis
            if (s.Position.Y < s.Radius)
            {
                s.Position.Y = s.Radius;
                if (s.Velocity.Y < 0f)
                    s.Velocity.Y *= -1f;
            }
            else if (s.Position.Y > worldSize - s.Radius)
            {
                s.Position.Y = worldSize - s.Radius;
                if (s.Velocity.Y > 0f)
                    s.Velocity.Y *= -1f;
            }

            // And lastly the Z axis
            if (s.Position.Z < -worldSize + s.Radius)
            {
                s.Position.Z = -worldSize + s.Radius;
                if (s.Velocity.Z < 0f)
                    s.Velocity.Z *= -1f;
            }
            else if (s.Position.Z > worldSize - s.Radius)
            {
                s.Position.Z = worldSize - s.Radius;
                if (s.Velocity.Z > 0f)
                    s.Velocity.Z *= -1f;
            }
        }

        #region HandleInput
        protected void HandleInput()
        {
            prevGamePadState = currentGamePadState;
            currentGamePadState = GamePad.GetState(PlayerIndex.One);
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState(PlayerIndex.One);

            // Check for exit.
            if (IsPressed(Keys.Escape, Buttons.Back))
            {
                Exit();
            }

            // If the X key or button was pressed, toggle whether or not we're colliding the spheres
            if (IsPressed(Keys.X, Buttons.X))
            {
                collideSpheres = !collideSpheres;
            }

            // If the user pressed Up or Down on the keyboard, DPad, or left thumbstick, we adjust
            // the number of active spheres in our scene
            if (IsPressed(Keys.Up, Buttons.DPadUp)) 
            {
                activeSphereCount++;
            }
            else if (IsPressed(Keys.Down, Buttons.DPadDown))
            {
                activeSphereCount--;
            }
        }

        /// <summary>
        /// Checks whether the specified key or button has been pressed.
        /// </summary>
        bool IsPressed(Keys key, Buttons button)
        {
            return (currentKeyboardState.IsKeyDown(key) &&
                    prevKeyboardState.IsKeyUp(key)) ||
                   (currentGamePadState.IsButtonDown(button) &&
                    prevGamePadState.IsButtonUp(button));
        }
        #endregion

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // TODO: Add your drawing code here

            // Create a view and projection matrix for our camera
            Matrix view = Matrix.CreateLookAt(
                new Vector3(worldSize, worldSize, worldSize) * 1.5f, Vector3.Zero, Vector3.Up);
            Matrix projection = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.PiOver4, GraphicsDevice.Viewport.AspectRatio, 0.01f, 100f);

            // Draw all of our spheres
            for (int i = 0; i < activeSphereCount; i++)
            {
                spheres[i].Draw(view, projection);
            }

            // Draw overlay text.
            string text = "X = Toggle Collision On/Off\n" +
                          "Up/Down = Increase/Decrease Number of Spheres\n";
            string demoText = string.Format(
                "Sphere count: {0}\nCollisions Enabled: {1}\n\n{2}",
                activeSphereCount, collideSpheres, text);            
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, demoText, new Vector2(48, 48), Color.White);
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
